/*
   lac -- a lisp interpreter
   Copyright (C) 2010 Gianluca Guida

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

%{
#define YYSTYPE lreg_t

#include <ctype.h>
#include <strings.h>

#include <lac.h>
#include "sexpr.h"
#include "private.h"
%}

%option noyywrap
%option reentrant
%option bison-bridge

WHITESPACE [\ \t\n]
DIGIT [0-9]
HEXDIGIT [0-9a-fA-F]
ALPHA [A-Za-z]
SP_INIT [#\*/:<=>?^_~!$%&+-]
SPECIAL [\*/:<=>?^_~!$%&+-.] 
SIGN [+|-]

%%

;(.)*
{WHITESPACE}

{SIGN}?{DIGIT}+ {
	intptr_t n;

	errno = 0;
	n = strtol(yytext, (char **)NULL, 10);
        *yylval = lac_extty_box(LREG_INTEGER, (void *)n, 0);
	return INTEGER;
}

(#x|0x){HEXDIGIT}+ {
	intptr_t n;

	errno = 0;
	n = strtol(yytext+2, (char **)NULL, 16);
	if ( errno == ERANGE )
		raise_exception("Integer overflow in input", NIL);
	*yylval = lac_extty_box(LREG_INTEGER, (void *)n, 0);
	return INTEGER;
 }

(NIL|nil) {
	*yylval = NIL;
	return NIHIL;
 }

\"[^\"]*\" {
	size_t len = strlen(yytext);
	char *s = lac_alloc(len - 1);
	memcpy(s, yytext+1, len - 2);
	*yylval = lreg_raw(s, LREG_STRING);
	return STRING;
 }

(({DIGIT}+({SPECIAL}|{ALPHA}))|{ALPHA}|{SP_INIT})({SPECIAL}|{ALPHA}|{DIGIT})* {
	int i;
	size_t len = strlen(yytext);
	char *s = lac_alloc(len + 1);
	for (i = 0; i < len; i++)
		*(s+i) = toupper((int)*(yytext+i));
	*yylval = intern_symbol(s);
	return SYMBOL;
 }

,@ { return COMMA_AT; }

\'  { return QUOTE; }

` { return QUASI; }

, { return COMMA; }

\( { return OPAREN; }

\) { return CPAREN; }

\.  { return DOT; }

<<EOF>> { return -1; }

. { return -1; }


%%

static void _sexpr_parse(yyscan_t yyscan, struct parse_cb *cb)
{
	int v;
	lreg_t lr;
	void *parser;

	parser = ParseAlloc(malloc);

	lac_on_error({
			ParseFree(parser, free);
			yylex_destroy(yyscan);
			_throw();
	  });

	while((v = yylex(&lr, yyscan)) != -1) {
		Parse(parser, v, lr, cb);
	}
	Parse(parser, ENDPARSE, lr, cb);

	lac_off_error();

	ParseFree(parser, free);
	yylex_destroy(yyscan);
}

lreg_t sexpr_eval_string(char *s, lenv_t *env)
{
	yyscan_t yyscan;
	struct parse_cb cb;

	cb.fn = (lreg_t (*)(lreg_t,void*))eval;
	cb.opq = (void *)env;
	cb.res = NIL;

	yylex_init(&yyscan);
	yy_scan_string(s, yyscan);

	lac_on_error ({
	    lac_error_print (stderr);
	    return NIL;
	  });

	_sexpr_parse(yyscan, &cb);

	lac_off_error();
	return cb.res;
}

lreg_t sexpr_parse_string(char *s, lreg_t (*fn)(lreg_t,void*), void *opq)
{
	yyscan_t yyscan;
	struct parse_cb cb;

	cb.fn = fn;
	cb.opq = opq;
	cb.res = NIL;

	yylex_init(&yyscan);
	yy_scan_string(s, yyscan);

	_sexpr_parse(yyscan, &cb);

	return cb.res;
}

lreg_t sexpr_parse_file(FILE *f, lreg_t (*fn)(lreg_t,void*), void *opq)
{
	yyscan_t yyscan;
	struct parse_cb cb;

	cb.fn = fn;
	cb.opq = opq;
	cb.res = NIL;

	yylex_init(&yyscan);
	yyset_in(f, yyscan);

	_sexpr_parse(yyscan, &cb);
	return cb.res;
}


/*
 * Print
 */

static void sexpr_print_cons(FILE *f, lreg_t lr, const char *h)
{
  lreg_t a = car(lr);
  lreg_t d = cdr(lr);

  printf(h);
  sexpr_fprint(f, a);

  if (d == NIL) {
    fprintf(f, ")");
    return;
  }

  if (lreg_raw_type(d) == LREG_CONS)
    sexpr_print_cons(f, d, " ");
  else
    {
      fprintf(f, " . ");
      sexpr_fprint(f, cdr(lr));
      fprintf(f, ")");
    }
}

void sexpr_fprint(FILE *f, lreg_t lr)
{
  switch ( lreg_type(lr) )
    {
    case LREG_NIL:
      fprintf(f, "()");
      break;
    case LREG_SYMBOL:
      fprintf(f, "%s", (char *)lreg_raw_ptr(lr));
      break;
    case LREG_STRING:
      fprintf(f, "\"%s\"", (char *)lreg_raw_ptr(lr));
      break;
    case LREG_CONS:
      sexpr_print_cons(f, lr, "(");
      break;
    case LREG_MACRO:
      fprintf(f, "<#MACRO>");
      break;
    case LREG_LAMBDA:
      fprintf(f, "<#LAMBDA>");
      break;
    case LREG_LLPROC:
      fprintf(f, "<#LLPROC>");
      break;
    default:
      if ( !lac_extty_print(f, lr) )
	fprintf(f, "<??? %d>",(int)lreg_type(lr));
    }
  return;
}
